import abc
import logging


class IBaseReader(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def read(self, *args, **kwargs):
        pass


class TextFileReader(IBaseReader):

    def __init__(self, *args, **kwargs):
        self.filename = kwargs.get(u'filename')

    def read(self, *args, **kwargs):
        try:
            with open(self.filename, u'r') as f:
                return iter(f.read().splitlines())
        except IOError:
            logging.error(u"Can't read file: {}.".format(self.filename))
            return []



import abc
import logging
import random
import time

import tornado.gen
import tornado.ioloop
import tornado.locks
from tornado.httpclient import AsyncHTTPClient, HTTPError

from . import exceptions
from . import readers
from . import utils
from . import writers


HTTP_MAX_CLIENTS = 1000
CONNECT_TIMEOUT = 10
REQUEST_TIMEOUT = 1000

READER_MAP = {
    utils.FILE_READER: readers.TextFileReader,
}

WRITER_MAP = {
    utils.MEMORY_WRITER: writers.MemoryWriter,
    utils.REDIS_WRITER: writers.RedisWriter
}


class IBaseMonitor(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, reader=u'file', writer=u'memory', *args, **kwargs):
        self.reader = reader
        self.writer = writer
        self.client = None
        self.primitive = None
        self.reader_instance = None
        self.writer_instance = None
        self.concurrency = kwargs.get(u'concurrency', 0)
        self.reader_map = kwargs.get(u'reader_map', READER_MAP)
        self.writer_map = kwargs.get(u'writer_map', WRITER_MAP)

    def get_concurrency_primitive(self):
        return tornado.locks.Semaphore(self.concurrency)

    def get_reader_instance(self):
        try:
            ReaderClass = self.reader_map[self.reader]
            return ReaderClass(**self.kwargs)
        except KeyError:
            msg = u"Inappropriate reader. Allowed: {0}".format(
                u", ".join(self.reader_map.iterkeys())
            )
            raise exceptions.ConfigurationException(msg)

    def get_writer_instance(self):
        try:
            WriterClass = self.writer_map[self.writer]
            return WriterClass(**self.kwargs)
        except KeyError:
            msg = u"Inappropriate writer. Allowed: {0}".format(
                u", ".join(self.reader_map.iterkeys())
            )
            raise exceptions.ConfigurationException(msg)

    def setup(self):
        if self.concurrency:
            logging.info(u"Concurrency set to {0}".format(self.concurrency))
            self.primitive = self.get_concurrency_primitive()

        self.reader_instance = self.get_reader_instance()
        self.writer_instance = self.get_writer_instance()
        self.client = self.get_client_instance()

    @abc.abstractmethod
    def monitor(self, url):
        pass

    def get_client_instance(self):
        return AsyncHTTPClient(max_clients=HTTP_MAX_CLIENTS)

    @tornado.gen.coroutine
    def iter_urls(self, *args, **kwargs):
        data = yield tornado.gen.Task(self.writer_instance.iter_data)
        raise tornado.gen.Return(data)

    @tornado.gen.coroutine
    def get_info(self, url):
        data = yield tornado.gen.Task(self.writer_instance.get_info, url)
        raise tornado.gen.Return(data)

    @tornado.gen.coroutine
    def get_stats(self):
        data = yield tornado.gen.Task(self.writer_instance.get_stats)
        raise tornado.gen.Return(data)

    def start(self, *args, **kwargs):
        self.setup()

        for url in self.reader_instance.read():
            logging.info(u'{} - starting monitoring'.format(url))
            self.monitor(url=url)

    def stop(self, *args, **kwargs):
        logging.info(u"Stopping {}.".format(self.__class__.__name__))


class WebMonitor(IBaseMonitor):

    def __init__(self, *args, **kwargs):
        super(WebMonitor, self).__init__(*args, **kwargs)
        self.kwargs = kwargs

    @tornado.gen.coroutine
    def monitor_url(self, url):
        now = time.time()
        try:
            response = yield self.client.fetch(
                url, method=u'HEAD',
                request_timeout=REQUEST_TIMEOUT,
                connect_timeout=CONNECT_TIMEOUT
            )
            self.writer_instance.write_response(url=url, response=response)
        except HTTPError as e:
            logging.error(u"[{}] {}".format(url, str(e)))
            self.writer_instance.write_error(url=url, error=e)

        deadline = now + random.randint(5, 15)

        callback = lambda: self.monitor(url=url)
        tornado.ioloop.IOLoop.instance().add_timeout(
            deadline=deadline,
            callback=callback
        )

    @tornado.gen.coroutine
    def monitor(self, url):
        if self.primitive:
            with (yield self.primitive.acquire()):
                self.monitor_url(url=url)
        else:
            self.monitor_url(url=url)


import os
import time
import signal
import logging

import tornado.ioloop
import tornado.web
from tornado.ioloop import IOLoop
from tornado.httpserver import HTTPServer
from tornado.options import define, options
from tornado.httpclient import AsyncHTTPClient

from tormon.monitor import client
from tormon.api import urls as api_urls
from tormon.web import urls as web_urls


root = lambda *x: os.path.join(os.path.dirname(__file__), u'../', *x)
APP_ROOT = root()
STATIC_ROOT = root(u'static')
TEMPLATE_ROOT = root(u'templates')
BLOCKING_THRESHOLD = 0.2

AsyncHTTPClient.configure(u"tornado.curl_httpclient.CurlAsyncHTTPClient")

define(u"reader", default=u"file", help=u"Name of reader that's used to provide urls")
define(u"writer", default=u"memory", help=u"Name of writer that's persisting urls")
define(u'filename', default=None, help=u"Path to file, required by file reader")
define(u"host", default=u"localhost")
define(u"port", default=8081, type=int)
define(u"debug", default=True, type=bool)
define(u"concurrency", default=0, type=int)


class MonitorWebApplication(tornado.web.Application):

    def __init__(self, monitor_instance, **kwargs):
        handlers = api_urls.urlpatterns + web_urls.urlpatterns
        self.monitor = monitor_instance
        tornado.web.Application.__init__(self, handlers, **kwargs)


def shutdown(server_instance, monitor_instance):
    ioloop_instance = IOLoop.instance()
    logging.info(u'Stopping APP-MONITOR gracefully.')

    server_instance.stop()
    monitor_instance.stop()

    def finalize():
        ioloop_instance.stop()
        logging.info(u'APP-MONITOR stopped.')

    ioloop_instance.add_timeout(time.time() + 1.5, finalize)


def main():
    options.parse_command_line()

    monitor = client.WebMonitor(
        reader=options[u'reader'],
        writer=options[u'writer'],
        concurrency=options[u'concurrency'],
        filename=options[u'filename'],
    )
    application = MonitorWebApplication(
        monitor_instance=monitor,
        debug=options[u'debug'],
        static_path=STATIC_ROOT,
        template_path=TEMPLATE_ROOT
    )
    server = HTTPServer(application)

    logging.info(u"Starting APP-MONITOR on {0}:{1}.".format(
        options[u'host'], options[u'port']
    ))

    shutdown_handler = lambda sig, frame: shutdown(server, monitor)
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)

    if options[u'debug']:
        logging.info(u"Setting blocking threshold to: {}".format(
            BLOCKING_THRESHOLD
        ))
        IOLoop.instance().set_blocking_log_threshold(BLOCKING_THRESHOLD)

    server.listen(options[u'port'], options[u'host'])
    monitor.start()
    IOLoop.instance().start()


if __name__ == '__main__':
    main()